"""
Filter out of bounds images.
"""

from typing import List
from shapely.geometry import Polygon
from ..data import SatelliteImage


def filter_oob(
    satellite_images: List[SatelliteImage],
    polygon_geometry: Polygon,
    crs: str,
) -> List[SatelliteImage]:
    """
    Filter out images that are not within the bounds of a box.

    Args:
        satellite_images (List[SatelliteImage]): List of satellite images.
        polygon_geometry (Polygon): Polygon.
        crs (str): EPSG of the bounds.

    Returns:
        List[SatelliteImage]: List of satellite images within the bounds of the box.
    """
    inbound_images = []
    for satellite_image in satellite_images:
        if satellite_image.intersects_polygon(
            polygon_geometry,
            crs,
        ):
            inbound_images.append(satellite_image)
    return inbound_images



"""
Data module.
"""

from .satellite_image import SatelliteImage
from .labeled_satellite_image import (
    SegmentationLabeledSatelliteImage,
    DetectionLabeledSatelliteImage,
    ClassificationLabeledSatelliteImage,
)

__all__ = [
    "SatelliteImage",
    "SegmentationLabeledSatelliteImage",
    "DetectionLabeledSatelliteImage",
    "ClassificationLabeledSatelliteImage",
]



"""
Plot module.
"""

from .plot_utils import (
    plot_images,
    plot_images_with_segmentation_label,
    plot_images_with_classification_label,
    plot_images_with_detection_label,
    make_mosaic,
)

__all__ = [
    "plot_images",
    "plot_images_with_segmentation_label",
    "plot_images_with_classification_label",
    "plot_images_with_detection_label",
    "make_mosaic",
]



"""
Cloud filtering.
"""

from typing import List
import numpy as np
from ..data import SatelliteImage


def filter_cloudy(
    satellite_images: List[SatelliteImage],
    cloud_masks: List[np.array],
    cloud_threshold: float = 0.5,
):
    """
    Filter out cloudy images.
    """
    if len(satellite_images) != len(cloud_masks):
        raise ValueError("Length of satellite_images and cloud_masks must be the same.")
    cloudless_images = []
    for satellite_image, cloud_mask in zip(satellite_images, cloud_masks):
        cloud_coverage = np.sum(cloud_mask) / np.prod(cloud_mask.shape)
        if cloud_coverage < cloud_threshold:
            cloudless_images.append(satellite_image)
    return cloudless_images



"""
Sample module.
"""

from .sample import (
    compute_distance_to_point,
    is_within_distance,
    sample_around_coordinates,
)


__all__ = [
    "compute_distance_to_point",
    "is_within_distance",
    "sample_around_coordinates",
]



"""
Functions to sample from a list of satellite images.
"""

from geopy.distance import geodesic
from pyproj import Transformer
from typing import List, Tuple
from ..data import SatelliteImage


def compute_distance_to_point(
    satellite_image: SatelliteImage, coordinates: List[float], crs: str = "EPSG:4326"
) -> float:
    """
    Return distance between the centroid of a satellite image
    and a point with coordinates `coordinates`.

    Args:
        satellite_image (SatelliteImage): Satellite image.
        coordinates (List[float]): Coordinates of the point.
        crs (str): EPSG of coordinates. Defaults to "EPSG:4326".

    Returns:
        float: Distance in kilometers.
    """
    # Reproject coordinates to lat and long if needed
    if crs != "EPSG:4326":
        lat, lon = project(coordinates, crs, "EPSG:4326")
    else:
        lat, lon = coordinates

    # Get centroid coordinates
    x0, y0, x1, y1 = satellite_image.bounds
    centroid_x = (x0 + x1) / 2
    centroid_y = (y0 + y1) / 2

    # Reproject centroid coordinates if needed
    if satellite_image.crs != "EPSG:4326":
        centroid_lat, centroid_lon = project(
            (centroid_x, centroid_y), satellite_image.crs, "EPSG:4326"
        )
    else:
        centroid_lat, centroid_lon = centroid_x, centroid_y

    # Return the geodesic distance between the two points
    return geodesic((lat, lon), (centroid_lat, centroid_lon)).meters / 1000


def is_within_distance(
    satellite_image: SatelliteImage,
    distance: float,
    coordinates: List[float],
    crs: str = "EPSG:4326",
) -> bool:
    """
    Return True if the centroid of a satellite image is within distance
    `distance` of a point with coordinates `coordinates`.

    Args:
        satellite_image (SatelliteImage): Satellite image.
        distance (float): Distance in kilometers.
        coordinates (List[float]): Coordinates of the point.
        crs (str): EPSG. Defaults to "EPSG:4326".

    Returns:
        bool: True if image is within distance.
    """
    if (
        compute_distance_to_point(
            satellite_image=satellite_image, coordinates=coordinates, crs=crs
        )
        < distance
    ):
        return True
    return False


def sample_around_coordinates(
    satellite_images: List[SatelliteImage],
    distance: float,
    coordinates: List[float],
    crs: str = "EPSG:4326",
) -> Tuple[List[SatelliteImage]]:
    """
    Sample satellite images which are within a distance
    `distance` from a point with coordinates `coordinates`.

    Args:
        satellite_image (SatelliteImage): Satellite image.
        distance (float): Distance in kilometers.
        coordinates (List[float]): Coordinates of the point.
        crs (str): EPSG. Defaults to "EPSG:4326".

    Returns:
        Tuple[List[SatelliteImage]]: Tuple containing two lists,
            the first containing the images which are within the
            given distance of the specified point and the second
            containing the others.
    """
    inside_images = []
    outside_images = []

    for satellite_image in satellite_images:
        if is_within_distance(
            satellite_image=satellite_image,
            distance=distance,
            coordinates=coordinates,
            crs=crs,
        ):
            inside_images.append(satellite_image)
        else:
            outside_images.append(satellite_image)
    return inside_images, outside_images


def project(coordinates: List[float], input_crs: str, output_crs: str) -> List[float]:
    """
    Reproject coordinates.

    Args:
        coordinates (List[float]): Coordinates.
        input_crs (str): EPSG.
        output_crs (str): EPSG.

    Returns:
        List[float]: Reprojected coordinates.
    """
    transformer = Transformer.from_crs(input_crs, output_crs)
    input_lat, input_lon = coordinates
    return transformer.transform(*coordinates)



"""
Filtering functions.
"""

import numpy as np
from typing import List
from ..data import SatelliteImage


def is_corrupted(
    image: SatelliteImage,
    black_value_threshold: int = 25,
    black_area_threshold: float = 0.5,
) -> bool:
    """
    Check if image is corrupted.

    Args:
        image (SatelliteImage): Satellite image.
        black_value_threshold (int): The intensity threshold to consider
            a pixel as black. Pixels with intensity values less than
            this threshold are considered black. Default is 100.
        black_area_threshold (float): The threshold for the proportion
            of black pixels in the image. If the ratio of black pixels
            exceeds this threshold, the function returns True. Default is 0.5.

    Returns:
        bool: True if image is corrupted, False otherwise.
    """
    # Convert the RGB image to grayscale
    gray_image = (
        0.2989 * image.array[0] + 0.5870 * image.array[1] + 0.1140 * image.array[2]
    )

    # Count the number of black pixels
    nb_black_pixels = np.sum(gray_image < black_value_threshold)

    # Calculate the proportion of black pixels
    black_pixel_ratio = nb_black_pixels / np.prod(gray_image.shape)

    # Check if the proportion exceeds the threshold
    return black_pixel_ratio >= black_area_threshold


def filter_corrupted(
    satellite_images: List[SatelliteImage],
    black_value_threshold: int = 25,
    black_area_threshold: float = 0.5,
) -> List[SatelliteImage]:
    """
    Filter out corrupted images.

    Args:
        satellite_images (List[SatelliteImage]): List of satellite images.
    Returns:
        List[SatelliteImage]: List of satellite images without corrupted images.
    """
    corrupted_images = []
    for satellite_image in satellite_images:
        if not is_corrupted(
            satellite_image, black_value_threshold, black_area_threshold
        ):
            corrupted_images.append(satellite_image)
    return corrupted_images
